"""Module handling scaling of the input variables

"""
import json
import numpy as np
import pandas as pd
import HelperModules.helperfunctions as hf
from HelperModules.messagehandler import WarningMessage


class scaler():
    """ Class to handle the scaling of the data

    """

    def __init__(self, cfgset):
        self.m_cfgset = cfgset
        self.m_data = self._getdata()
        self.m_recalc_scaler = self.m_cfgset.get("GENERAL").get("RecalcScaler")
        self.m_scale_value_dict = {}
        self._getscaled()

    def _getdata(self):
        """Load and return merged.h5

        :returns: merged.h5

        """
        return pd.read_hdf(f"{hf.addslash(self.m_cfgset.get('DIR').get('Data'))}merged.h5")

    def _load_scaler(self):
        """Load the already existing scaler

        :returns: Dictionary with Scaling:{Variables:{Offset:..., Scale:...}}

        """
        with open(f"{hf.addslash(self.m_cfgset.get('DIR').get('Model'))}scaler.json", "r", encoding="utf-8") as j:
            content = json.load(j)
        return content

    def _pT_scaling(self, groups):
        """pT_scale input variables

        :param groups: dictionary assigns each type a list of variables, which pT_scales together
        :returns:

        """
        for names in groups:
            group_variables = groups[names]
            if len(group_variables) == 1:
                WarningMessage(f"Only one variable in pT_scaling Group {names}. Variable {group_variables} Scaled to 0")
            factor = self.m_data[group_variables].sum(axis=1).values
            self.m_data[group_variables] = np.log(self.m_data[group_variables].divide(factor, axis=0))

    def _minmax_scaling(self, names, minimum=0, maximum=1):
        """minmax scale input variables and calculate the offset and scale values. The scaling is to be applied as follows: x_scaled = x * scale + offset

        :param names: list of variables, which are all minmax scaled
        :param minimum: Determines the lower interval value, to which the variables are scaled to
        :param maximum: Determines the upper interval value, to which the variables are scaled to
        :returns: dictionary with Variables:{"Offset":... , "Scale":... } pair

        """
        if not self.m_recalc_scaler:
            scales = self._load_scaler()
            if "minmax" in scales:
                for var in scales["minmax"]:
                    self.m_data[var] = self.m_data[var] * scales["minmax"][var]["Scale"] + scales["minmax"][var]["Scale"]
                return scales["minmax"]

            if "minmax_symmetric" in scales:
                for var in scales["minmax_symmetric"]:
                    self.m_data[var] = self.m_data[var] * scales["minmax_symmetric"][var]["Scale"] + scales["minmax_symmetrc"][var]["Offset"]
                return scales["minmax_symmetric"]

        max_values = self.m_data[names].max(axis=0).values
        min_values = self.m_data[names].min(axis=0).values
        max_min_diff = max_values - min_values

        minmax_scale_dict = {names[i]: {"Offset": minimum - min_values[i] * (maximum - minimum) / (max_min_diff[i]), "Scale": (maximum - minimum) / (max_min_diff[i])} for i in range(len(names))}
        self.m_data[names] = minimum + self.m_data[names].subtract(min_values, axis=1).divide(max_min_diff) * (maximum - minimum)
        return minmax_scale_dict

    def _standard_scaling(self, names):
        """Standard scale input variables and calculate the offset and scale values. The scaling is to be applied as follows: x_scaled = x * scale + offset

        :param names: list of variables, which are all standard scaled
        :returns: dictionary with Variables:{"Offset":... , "Scale":... } pair

        """
        if not self.m_recalc_scaler:
            scales = self._load_scaler()
            if "standard" in scales:
                for var in scales["standard"]:
                    self.m_data[var] = self.m_data[var] * scales["standard"][var]["Scale"] + scales["standard"][var]["Offset"]
                return scales["standard"]

        mean_values = self.m_data[names].mean(axis=0).values
        std_values = self.m_data[names].std(axis=0).values

        stand_scale_dict = {names[i]: {"Offset": mean_values[i] / std_values[i], "Scale": std_values[i]} for i in range(len(names))}
        self.m_data[names] = self.m_data[names].subtract(mean_values).divide(std_values)
        return stand_scale_dict

    def _getscaling(self):
        """Find for each scaling the variables, which belongs to that scaling

        :returns: Dictionary with Scaling:[Variables], except pT_scaling:([Variables],[Type]) pair

        """
        scale_dict = {}
        vnames = np.array([v.get("Name") for v in self.m_cfgset.get("VARIABLE")])
        vscales = np.array([v.get("Scaling") for v in self.m_cfgset.get("VARIABLE")])
        vtypes = np.array([v.get("Type") for v in self.m_cfgset.get("VARIABLE")])
        index = 0

        for opt in vscales:
            if opt is None:
                opt = self.m_cfgset.get("GENERAL").get("InputScaling")
            if opt in scale_dict:
                if opt == "pT_scaling":
                    scale_dict[opt][0].append(vnames[index])
                    scale_dict[opt][1].append(vtypes[index])
                else:
                    scale_dict[opt].append(vnames[index])
            else:
                if opt == "pT_scaling":
                    scale_dict[opt] = ([vnames[index]], [vtypes[index]])
                else:
                    scale_dict[opt] = [vnames[index]]
            index += 1
        return scale_dict

    def _getpT_scale_groups(self):
        """Find for each pT_scale type the corresponding variables

        :returns: Dictionary with Types:[Variables] pair

        """
        pT_groups = {}
        if "pT_scaling" in self._getscaling():
            pT_variables, pT_types = (self._getscaling())["pT_scaling"]
            unique_types = np.unique(pT_types)
            pT_types = np.array(pT_types)  # Conversion into array needed, so the masking works.
            pT_variables = np.array(pT_variables)
            for types in unique_types:
                group_mask = (pT_types == types)
                pT_groups[types] = pT_variables[group_mask]
        return pT_groups

    def _dataframe_cleaner(self):
        """Deletes all columns of the inital dataframe, which do not belong to a variable

        :returns:

        """
        vnames = np.array([v.get("Name") for v in self.m_cfgset.get("VARIABLE")])
        df_columns = np.array(self.m_data.columns.values)
        delete_columns = [names for names in df_columns if names not in vnames]
        self.m_data = self.m_data.drop(columns=delete_columns)

    def _getscaled(self):
        """Scaler, which calls all the scaling options and saves the offset/scale of the variables in a json file

        :returns:

        """
        self._dataframe_cleaner()
        scale_dict = self._getscaling()
        pT_groups = self._getpT_scale_groups()
        # Perform minmax scaling and put into scale_value_dict to safe offset and scaling value
        if "minmax" in scale_dict:
            self.m_scale_value_dict["minmax"] = self._minmax_scaling(scale_dict["minmax"])

        # Perform None scaling
        if "None" in scale_dict:
            none_variables = scale_dict["None"]
            WarningMessage(f"The Variables: {none_variables} are NOT scaled")

        # Perform minmax_symmetric scaling and put into scale_value_dict to safe offset and scaling value
        if "minmax_symmetric" in scale_dict:
            self.m_scale_value_dict["minmax_symmetric"] = self._minmax_scaling(scale_dict["minmax_symmetric"], -1, 1)

        # Perform pT_scaling
        if pT_groups:
            self._pT_scaling(pT_groups)

        # Perform standard scaling and put into scale_value_dict to safe offset and scaling value
        if "standard" in scale_dict:
            self.m_scale_value_dict["standard"] = self._standard_scaling(scale_dict["standard"])

        # Save the scaling information for standard and minmax scaling into a json file. File in which it will be saved is: f"{hf.addslash(self.m_cfgset.get('DIR').get('Model'))}scaler.pkl"
        if self.m_recalc_scaler:
            with open(f"{hf.addslash(self.m_cfgset.get('DIR').get('Model'))}scaler.json", 'w', encoding="utf-8") as json_file:
                json.dump(self.m_scale_value_dict, json_file)
